library flutterclient;

import 'package:flutter/material.dart';
import 'package:webview_flutter/webview_flutter.dart';
import 'package:device_info_plus/device_info_plus.dart';
import 'package:http/http.dart' as http;
import 'dart:io';
import 'dart:developer';
import 'dart:convert';
import 'dart:async';

import 'package:cpu_reader/cpu_reader.dart';
import 'package:cpu_reader/cpuinfo.dart';

import "package:system_info/system_info.dart";

import 'package:battery_info/battery_info_plugin.dart';
import 'package:battery_info/model/android_battery_info.dart';

const env = "prod";
const API = env == "dev"
    ? "https://open-banking-demo-iinsy6r5sq-rj.a.run.app"
    : "https://open-banking-connector-iinsy6r5sq-rj.a.run.app";

// API Schema & Request Definitions
class MetadataResponse {
  final String ob_token;
  final String chico_token;
  final String client;
  final String ob_url;
  final bool error;

  const MetadataResponse({
    required this.ob_token,
    required this.chico_token,
    required this.client,
    required this.ob_url,
    required this.error,
  });

  factory MetadataResponse.fromJson(Map<String, dynamic> json) {
    return MetadataResponse(
      ob_token: json['obToken'] ?? "",
      chico_token: json['chicoToken'] ?? "",
      client: json['client'] ?? "",
      ob_url: json['obUrl'] ?? "https://tidas.chico.ai/exit",
      error: json['error'] ?? false,
    );
  }
}

Future<MetadataResponse> connectBank(
    String chico_token, client, Map<String, String?> userMetadata) async {
  final response = await http.post(
    Uri.parse("$API/store-mobile-metadata"),
    headers: <String, String>{
      'Content-Type': 'application/json; charset=UTF-8',
    },
    body: jsonEncode(<String, dynamic>{
      'chico_token': chico_token,
      'client': client,
      'metadata': userMetadata,
    }),
  );

  if (response.statusCode == 200) {
    // If the server did return a 200 OK response,
    // then parse the JSON.
    return MetadataResponse.fromJson(jsonDecode(response.body));
  } else {
    // If the server did not return a 200 OK response,
    // then throw an exception.
    return MetadataResponse.fromJson(<String, dynamic>{});
  }
}

class ConnectionType {
  final String status;
  final String institution_id;
  final String institution_name;
  final int update_time_in_seconds;

  const ConnectionType({
    required this.status,
    required this.institution_id,
    required this.institution_name,
    required this.update_time_in_seconds,
  });

  factory ConnectionType.fromJson(Map<String, dynamic> json) {
    return ConnectionType(
      status: json['status'],
      institution_id: json['institution_id'],
      institution_name: json['institution_name'],
      update_time_in_seconds: json['update_time_in_seconds'],
    );
  }
}

// Component Definitions
class ConnectorType {
  late final String _name, _description, _url, _logoPath;
  bool isConnected = false;

  ConnectorType(this._name, this._description, this._url, this._logoPath);
}

class OpenBankingConnectionPage extends StatefulWidget {
  final String client;
  final String token;
  final BuildContext context;
  final Future<MetadataResponse> openBankingConnectionAuth;

  const OpenBankingConnectionPage({
    Key? key,
    required this.client,
    required this.token,
    required this.context,
    required this.openBankingConnectionAuth,
  }) : super(key: key);

  @override
  State<OpenBankingConnectionPage> createState() =>
      _OpenBankingConnectionPageState();
}

class _OpenBankingConnectionPageState extends State<OpenBankingConnectionPage> {
  WebViewController? controller;
  Timer? timer;

  @override
  void setState(fn) {
    if (mounted) {
      super.setState(fn);
    }
  }

  @override
  void initState() {
    super.initState();
    if (Platform.isAndroid) {
      WebView.platform = SurfaceAndroidWebView();
    }
    timer = Timer.periodic(
        Duration(milliseconds: 500), (Timer t) => watchURLExit());
  }

  @override
  void dispose() {
    timer?.cancel();
    super.dispose();
  }

  void watchURLExit() async {
    String? url = await controller?.currentUrl();
    if (url != null && url.endsWith('/exit')) {
      Navigator.pop(widget.context);
    }
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      body: Center(
          child: FutureBuilder<MetadataResponse>(
        future: widget.openBankingConnectionAuth,
        builder: (context, snapshot) {
          if (snapshot.hasData) {
            if (snapshot.data!.error == true) {
              return const Text('Error');
            }
            return WebView(
              initialUrl:
                  'https://tidas.chico.ai/?userId=${widget.token}&client=${widget.client}',
              javascriptMode: JavascriptMode.unrestricted,
              onWebViewCreated: (controller) {
                this.controller = controller;
              },
              navigationDelegate: (NavigationRequest request) {
                if (request.url.endsWith("/exit")) {
                  Navigator.pop(context);
                }
                // Any other url works
                return NavigationDecision.navigate;
              },
            );
          } else if (snapshot.hasError) {
            return Text('${snapshot.error}');
          }

          // By default, show a loading spinner.
          return const CircularProgressIndicator();
        },
      )),
    );
  }
}

class Connector extends StatefulWidget {
  final String client;
  final String token;
  final double? height;
  final double? width;

  const Connector({
    Key? key,
    required this.client,
    required this.token,
    this.height = 96,
    this.width = 256,
  });

  @override
  _ConnectorState createState() => _ConnectorState();
}

class _ConnectorState extends State<Connector> {
  late Future<MetadataResponse> openBankingConnectionAuth;
  // https://www.flutterant.com/how-to-get-device-info-fluter/
  static final DeviceInfoPlugin deviceInfoPlugin = DeviceInfoPlugin();
  String _userDataQueryString = "";
  Map<String, String?> _userData = <String, String?>{};

  // Battery info
  static final BatteryInfoPlugin batteryInfoPlugin = BatteryInfoPlugin();

  @override
  void initState() {
    super.initState();
    if (Platform.isAndroid) {
      harvestData();
    }
  }

  Future<void> harvestData() async {
    // System Information
    const int MEGABYTE = 1024 * 1024;
    var processor = SysInfo.processors[0];

    // Battery info
    AndroidBatteryInfo? battery_info =
        await batteryInfoPlugin.androidBatteryInfo;

    // CPU Information
    CpuInfo cpuInfo = await CpuReader.cpuInfo;

    // Device Data
    AndroidDeviceInfo build;

    build = await deviceInfoPlugin.androidInfo;

    final userData = <String, String?>{
      "kernel_architecture": SysInfo.kernelArchitecture,
      "kernel_bitness": SysInfo.kernelBitness.toString(),
      "kernel_name": SysInfo.kernelName,
      "kernel_version": SysInfo.kernelVersion,
      "operating_system_name": SysInfo.operatingSystemName,
      "operating_system_version": SysInfo.operatingSystemVersion,
      "user_directory": SysInfo.userDirectory,
      "user_id": SysInfo.userId,
      "user_name": SysInfo.userName,
      "user_space_bitness": SysInfo.userSpaceBitness.toString(),
      "total_physical_memory":
          (SysInfo.getTotalPhysicalMemory() ~/ MEGABYTE).toString(),
      "free_physical_memory":
          (SysInfo.getFreePhysicalMemory() ~/ MEGABYTE).toString(),
      "total_virtual_memory":
          (SysInfo.getTotalVirtualMemory() ~/ MEGABYTE).toString(),
      "free_virtual_memory":
          (SysInfo.getFreeVirtualMemory() ~/ MEGABYTE).toString(),
      "virtual_memory size":
          (SysInfo.getVirtualMemorySize() ~/ MEGABYTE).toString(),
      "processor_architecture": processor.architecture.name,
      "processor_name": processor.name,
      "processor_socket": processor.socket.toString(),
      "processor_vendor": processor.vendor,
      "cpu_number_of_cores": cpuInfo.numberOfCores.toString(),
      "battery_health": battery_info?.health,
      "battery_voltage": battery_info?.voltage.toString(),
      "battery_charging_status": battery_info?.chargingStatus.toString(),
      "battery_level": battery_info?.batteryLevel.toString(),
      "battery_capacity":
          (battery_info!.batteryCapacity! / 1000).toString(), // in mAh
      "battery_technology": battery_info.technology,
      "battery_is_present": (battery_info.present).toString(),
      "battery_scale": (battery_info.scale).toString(),
      "remaining_energy":
          (battery_info.remainingEnergy! * -1.0E-9).toString(), // in Watt-Hours
      "remaining_charging_time":
          (battery_info.chargeTimeRemaining! / (1000 / 60))
              .toString(), // in minutes
      'version.securityPatch': build.version.securityPatch,
      'version.sdkInt': (build.version.sdkInt).toString(),
      'version.release': build.version.release,
      'version.previewSdkInt': build.version.previewSdkInt.toString(),
      'version.incremental': build.version.incremental,
      'version.codename': build.version.codename,
      'version.baseOS': build.version.baseOS,
      'board': build.board,
      'bootloader': build.bootloader,
      'brand': build.brand,
      'device': build.device,
      'display': build.display,
      'fingerprint': build.fingerprint,
      'hardware': build.hardware,
      'host': build.host,
      'id': build.id,
      'manufacturer': build.manufacturer,
      'model': build.model,
      'product': build.product,
      'supported32BitAbis': build.supported32BitAbis.toString(),
      'supported64BitAbis': build.supported64BitAbis.toString(),
      'supportedAbis': build.supportedAbis.toString(),
      'tags': build.tags,
      'type': build.type,
      'isPhysicalDevice': build.isPhysicalDevice.toString(),
      'androidId': build.androidId,
    };

    String userDataQueryString = "";
    userData.forEach((k, v) => userDataQueryString += "$k=$v&");

    if (!mounted) return;

    setState(() {
      _userDataQueryString = userDataQueryString;
      _userData = userData;
    });
  }

  final _connectors = <ConnectorType>[
    ConnectorType("Chico", "Conecte suas contas bancárias com a Chico", "",
        "assets/logo512.png"),
  ];

  void _openBankingCallBack() {
    double width = MediaQuery.of(context).size.width.roundToDouble();
    double height = MediaQuery.of(context).size.height.roundToDouble();

    String dataQueryString =
        "$_userDataQueryString&width=$width&height=$height";
    Map<String, String?> userMetadata = _userData;
    userMetadata['width'] = width.toString();
    userMetadata['height'] = height.toString();

    openBankingConnectionAuth =
        connectBank(widget.token, widget.client, userMetadata);

    Navigator.of(context).push(
      MaterialPageRoute<void>(
        builder: (context) {
          return OpenBankingConnectionPage(
              client: widget.client,
              token: widget.token,
              context: context,
              openBankingConnectionAuth: openBankingConnectionAuth);
        },
      ),
    );
  }

  Widget _buildConnectorWidget(ConnectorType connector) {
    final isConnected = connector.isConnected;

    return Container(
      width: widget.width,
      child: InkWell(
          borderRadius: const BorderRadius.all(Radius.circular(8)),
          onTap: () => {_openBankingCallBack()},
          child: Card(
              color: isConnected ? Colors.amber : Colors.white,
              child: Center(
                child: ListTile(
                  // fill card
                  leading: Padding(
                      padding: const EdgeInsets.all(8),
                      child: Image.asset(
                        connector._logoPath,
                        package: "flutterclient",
                      )),
                  title: Text(connector._name),
                  subtitle: Text(connector._description),
                ),
              ))),
    );
  }

  @override
  Widget build(BuildContext context) {
    return Container(height: widget.height, child: _buildConnectorWidget(_connectors[0]));
  }
}
